V8 JavaScript Engine
=============

V8 is Google's open source JavaScript engine.

V8 implements ECMAScript as specified in ECMA-262.

V8 is written in C++ and is used in Google Chrome, the open source
browser from Google.

V8 can run standalone, or can be embedded into any C++ application.

V8 Project page: https://v8.dev/docs


Getting the Code
=============

Checkout [depot tools](http://www.chromium.org/developers/how-tos/install-depot-tools), and run

        fetch v8

This will checkout V8 into the directory `v8` and fetch all of its dependencies.
To stay up to date, run

        git pull origin
        gclient sync

For fetching all branches, add the following into your remote
configuration in `.git/config`:

        fetch = +refs/branch-heads/*:refs/remotes/branch-heads/*
        fetch = +refs/tags/*:refs/tags/*


Contributing
=============

Please follow the instructions mentioned at
[v8.dev/docs/contribute](https://v8.dev/docs/contribute).


mu*t*able-specific Changes
=============

We introduce a new function `SetWasmInstanceRawMemory()` to set the underlying memory of a Wasm module instance.  This enables passing data between the embedder and the Wasm module with zero copy.


```cpp
void SetWasmInstanceRawMemory(Local<Object> wasmInstance, uint8_t *mem_start, size_t mem_size) {
  auto instance = i::Handle<i::WasmInstanceObject>::cast(Utils::OpenHandle(*wasmInstance));
  instance->SetRawMemory(mem_start, mem_size);
  auto isolate = reinterpret_cast<i::Isolate*>(wasmInstance->GetIsolate());
  std::unique_ptr<i::BackingStore> store = i::BackingStore::WrapAllocation(isolate, mem_start, mem_size,
                                                                           i::SharedFlag::kNotShared, false);
  instance->memory_object().array_buffer().Attach(std::move(store));
}

```

In addition, we need to disable bounds checks.
We cannot expect the embedder to augment the data by guard pages or implement other mechanisms employed by V8 to perform bounds checks.
